local addon = PvPTimer

-- Libraries
addon.Lib = {
	AceConfigRegistry = LibStub("AceConfigRegistry-3.0"),
	AceConfigDialog = LibStub("AceConfigDialog-3.0"),
	AceGUI = LibStub("AceGUI-3.0"),
	Candy = LibStub("LibCandyBar-3.0"),
	CDFrames = LibStub("CooldownFrames-1.0"),
	Media = LibStub("LibSharedMedia-3.0"),
}

-- global tables
addon.Anchors = {}
addon.Interface = {}
addon.Units = {}
addon.Specs = {}
addon.OptionsGenerated = false
-- addon.CopyHeaderSettings = nil
-- addon.CopyBarSettings = nil

-- copy often used stuff to local variables
local L = addon.Locale
local cdf = addon.Lib.CDFrames
local media = addon.Lib.Media
local ACR = addon.Lib.AceConfigRegistry
local ACD = addon.Lib.AceConfigDialog

local units = addon.Units
local specs = addon.Specs
local defaults = addon.Defaults
local CopyTable = addon.CopyTable

local type = type
local pairs = pairs
local bband = bit.band
local smatch = string.match
local select = select

local UnitGUID, UnitName, UnitIsEnemy = UnitGUID, UnitName, UnitIsEnemy
local GetPlayerInfoByGUID = GetPlayerInfoByGUID
local GetTime = GetTime
local GetSpellInfo = GetSpellInfo
local GetBattlefieldStatus = GetBattlefieldStatus
local IsRatedBattleground = IsRatedBattleground
local IsInInstance = IsInInstance

-- events checked
local events = {
	["PLAYER_ENTERING_WORLD"] = true,
	["ZONE_CHANGED_NEW_AREA"] = true,
 	["COMBAT_LOG_EVENT_UNFILTERED"] = true,
	["PLAYER_TARGET_CHANGED"] = true,
	["PLAYER_FOCUS_CHANGED"] = true,
	["ARENA_OPPONENT_UPDATE"] = true,
}

-- combatlog event subtypes checked
local eventtypes = {
	["SPELL_CAST_SUCCESS"] = true,
}

local GUID_INVALID = "0x0000000000000000"

function addon:OnInitialize()
	--register chat commands
	addon:RegisterChatCommand("pvptimer", "ChatCommand")
	addon:RegisterChatCommand("pt", "ChatCommand")
	
	--create db
	addon.DB = LibStub("AceDB-3.0"):New("PTDB", defaults, true)
	local db = addon.DB

	--create options loader
	ACR:RegisterOptionsTable("PvPTimer", addon.Options)
	addon.OptionsFrame = ACD:AddToBlizOptions("PvPTimer", "PvPTimer", nil, "Settings")

	-- local build = tonumber(select(2, GetBuildInfo()))

	-- create default anchors
	local anchors = addon.Anchors
	anchors["Anchor_Target"] = cdf:CreateAnchor("pt_target", "Anchor_Target")
	anchors["Anchor_Focus"] = cdf:CreateAnchor("pt_focus", "Anchor_Focus")
	anchors["Anchor_Arena1"] = cdf:CreateAnchor("pt_arena1", "Anchor_Arena1")
	anchors["Anchor_Arena2"] = cdf:CreateAnchor("pt_arena2", "Anchor_Arena2")
	anchors["Anchor_Arena3"] = cdf:CreateAnchor("pt_arena3", "Anchor_Arena3")
	anchors["Anchor_Arena4"] = cdf:CreateAnchor("pt_arena4", "Anchor_Arena4")
	anchors["Anchor_Arena5"] = cdf:CreateAnchor("pt_arena5", "Anchor_Arena5")
	
	-- all done
	addon:Print(L["Addon loaded."])
end

function addon:OnEnable()
	local db = addon.DB
	local anchors = addon.Anchors

	--copy any missing options from default settings
	for a in pairs(anchors) do
		if smatch(a, "_Arena") then a = "Anchor_Arena" end
		for k, v in pairs(db.global.anchor_default) do
			if db.profile.Anchors[a][k] == nil or db.profile.Anchors[a][k] == {} then db.profile.Anchors[a][k] = addon:CopyTable(v) end
		end
	end

	-- update anchors
	addon:PLAYER_TARGET_CHANGED()
	addon:PLAYER_FOCUS_CHANGED()
	addon:ZONE_CHANGED_NEW_AREA()
	addon:ARENA_OPPONENT_UPDATE()
end

function addon:ChatCommand(input)
	if input == "sc" or input == "spell" or input == "spellconfig" then
		-- open spell configuration
		addon:ToggleSpellConfig()
	else
		-- open blizzard config
		InterfaceOptionsFrame_OpenToCategory(addon.OptionsFrame)
	end
end

--create options on demand
function addon:CreateBlizOptions()
	if addon.OptionsGenerated then return end
	
	addon.OptionsGenerated = true
	local db = addon.DB
	
	addon.Profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(db)

	addon.DB.RegisterCallback(addon, "OnProfileChanged", "OnEnable")
	addon.DB.RegisterCallback(addon, "OnProfileCopied", "OnEnable")
	addon.DB.RegisterCallback(addon, "OnProfileReset", "OnEnable")
	
	ACD:AddToBlizOptions("PvPTimer", "Anchor Settings", "PvPTimer", "Anchors")
	ACD:AddToBlizOptions("PvPTimer", "Alert Messages", "PvPTimer", "Alert")
	ACD:AddToBlizOptions("PvPTimer", "Spell Categories", "PvPTimer", "SpellCategories")

	-- local LibDualSpec = LibStub("LibDualSpec-1.0")
	-- LibDualSpec:EnhanceOptions(addon.Profiles, addon.DB)
	ACR:RegisterOptionsTable("PvPTimer-Profiles", addon.Profiles)
	ACD:AddToBlizOptions("PvPTimer-Profiles", "Profiles", "PvPTimer")
end

-- Refresh all anchors
function addon:RefreshAnchors()
	local db = addon.DB.profile.Anchors
	local anchors = addon.Anchors
	local zone = GetCurrentMapAreaID()
	local zonetype = select(2, IsInInstance())

	-- move, size, style anchors
	for k, v in pairs(anchors) do
		if not smatch(k, "_Arena") then
			local a = db[k]
			
			local ap = a.Position
			cdf:MoveAnchor(v, ap.AttachPoint, ap.AttachFrame, ap.AttachRelative, ap.AttachX, ap.AttachY)
			cdf:ResizeAnchor(v, a.Width, 16, a.Scale)
			local ah = a.Header
			cdf:SetHeaderStyle(v, ah.Texture, ah.Font, ah.FontSize, ah.Justify)
			local ab = a.Bars
			cdf:SetBarStyle(v, ab.Texture, ab.ShowName, ab.ShowTime, ab.NameFont, ab.NameFontSize, ab.TimeFont, ab.TimeFontSize)
		end
	end
	
	-- same for all arena anchors
	for i = 1, 5 do
		local k = "Anchor_Arena"..i
		local v = anchors[k]
		local a = db["Anchor_Arena"]
		
		local ap = a.Position
		cdf:MoveAnchor(v, ap.AttachPoint, ap.AttachFrame, ap.AttachRelative, ap.AttachX+(i-1)*ap.NextX, ap.AttachY+(i-1)*ap.NextY)
		cdf:ResizeAnchor(v, a.Width, 16, a.Scale)
		local ah = a.Header
		cdf:SetHeaderStyle(v, ah.Texture, ah.Font, ah.FontSize, ah.Justify)
		local ab = a.Bars
		cdf:SetBarStyle(v, ab.Texture, ab.ShowName, ab.ShowTime, ab.NameFont, ab.NameFontSize, ab.TimeFont, ab.TimeFontSize)
	end
	
	local e_target, e_focus = true, true

	-- check current zone and hide anchors accordingly
	e_target = addon:GetZoneSetting("Anchors", "Anchor_Target")
	e_focus = addon:GetZoneSetting("Anchors", "Anchor_Focus")
	
	-- hide/show arena frames
	if zonetype == "arena" and db["Anchor_Arena"].Enabled then
		-- always show first 2 anchors
		anchors["Anchor_Arena1"]:Show()
		anchors["Anchor_Arena2"]:Show()
		local teamsize = select(6, GetBattlefieldStatus())
		if teamsize >= 3 then
			-- team is 3v3 or bigger, show 3rd anchor
			anchors["Anchor_Arena3"]:Show()
		end
		if teamsize == 5 then
			-- team is 5v5, show 4th and 5th anchors
			anchors["Anchor_Arena4"]:Show()
			anchors["Anchor_Arena5"]:Show()
		end
		addon:ARENA_OPPONENT_UPDATE()
	else
		-- we aren't in an arena or arena anchors are disabled, hide all
		for i = 1, 5 do
			local a = "Anchor_Arena"..i
			cdf:ClearAnchor(anchors[a])
			anchors[a]:Hide()
		end
	end

	-- show/hide target anchor
	if db["Anchor_Target"].Enabled and e_target then
		anchors["Anchor_Target"]:Show()
		addon:PLAYER_TARGET_CHANGED()
	else
		cdf:ClearAnchor(anchors["Anchor_Target"])
		anchors["Anchor_Target"]:Hide()
	end

	-- show/hide focus anchor
	if db["Anchor_Focus"].Enabled and e_focus then
		anchors["Anchor_Focus"]:Show()
		addon:PLAYER_FOCUS_CHANGED()
	else
		cdf:ClearAnchor(anchors["Anchor_Focus"])
		anchors["Anchor_Focus"]:Hide()
	end
end

-- updates all spells on an anchor
function addon:UpdateAnchor(anchor, spells, GUID)
	local a_id
	if smatch(anchor.ID, "_Arena") then
		a_id = "Anchor_Arena"
	else
		a_id = anchor.ID
	end
	local db = addon.DB.profile.Anchors[a_id]

	if not anchor:IsShown() then anchor:Show() end

	-- handle autohide settings
	if GUID == "hide" then
		cdf:ClearAnchor(anchor)
		anchor:Hide()
		return
	elseif GUID == "notarget" then
		cdf:ClearAnchor(anchor)
		cdf:SetHeader(anchor, "No Target", "_HIDE_", "_HIDE_", db.Header.Color)
		return
	elseif GUID == "sticky" then
		return
	end

	local _, class, _, _, _, name = GetPlayerInfoByGUID(GUID)

	-- get header color (class color or default)
	local color
	if (db.Header.ClassColor and class and RAID_CLASS_COLORS[class]) then
		color = RAID_CLASS_COLORS[class]
		color.a = db.Header.Color.a
	else
		color = db.Header.Color
	end
	
	-- get unit's spec
	local spec
	if units[GUID] and units[GUID].spec then
		spec = units[GUID].spec
	end

	-- hide class/spec icons if they're turned off in options
	if db.Header.ShowClass == false then
		class = "_HIDE_"
	end
	if db.Header.ShowSpec == false then
		spec = "_HIDE_"
	end

	-- set header info
	cdf:SetHeader(anchor, name, class, spec, color)
	
	anchor.columns = db.Bars.Columns or 1
	anchor.growup = db.Bars.GrowUp

	-- redraw spell timer bars
	cdf:ClearAnchor(anchor)
	if not spells then return end
	for k, v in pairs(spells) do
		if type(k) == "number" then
			spell = addon:GetSpell(k)
			if not spell then return end
			local spelltype = spell.type or 'misc'
			local barcolor = addon:GetBarColor(a_id, spelltype)
			local spec = spell.spec or 0
			local glyphed = addon:GetGlyphSettings(k, spec)
			local cd = addon:GetSpellCooldown(k, glyphed, spec)
			local duration = v + cd - GetTime()
			if duration>0 then
				cdf:UpdateSpell(anchor, k, duration, barcolor)
			end
		end
	end
end

function addon:SendAlert(event, src, dst, spell, spellType)
	if not event or not src or not spell then return end
	local db = addon.DB.profile.Alert
	if not db[spellType] then spellType = 'misc' end

	-- check if alerts are enabled for this spell type
	if db[spellType].Enabled == false then return end

	local srcName = select(6, GetPlayerInfoByGUID(src))
	local dstName
	if dst and dst ~= GUID_INVALID then dstName = select(6, GetPlayerInfoByGUID(dst)) end
	local name
	if spell and type(spell) == "number" then name = GetSpellInfo(spell) end
	local zonetype = select(2, IsInInstance())
	local rated = IsRatedBattleground()
	local target, onlytarget = false
	
	-- check current zone and send alert accordingly
	if zonetype == "pvp" and not rated then -- we're in a normal bg
		target = db[spellType].BG
		onlytarget = db[spellType].BG_OnlyTarget
	elseif zonetype == "pvp" and rated then -- we're in a rated bg
		target = db[spellType].RatedBG
		onlytarget = db[spellType].RatedBG_OnlyTarget
	elseif zonetype == "arena" then -- we're in an arena
		target = db[spellType].Arena
		onlytarget = db[spellType].Arena_OnlyTarget
	else -- we're somewhere else
		target = db[spellType].Else
		onlytarget = db[spellType].Else_OnlyTarget
	end
	
	-- no info or alert target is 'none'
	if target == '_none_' then return end
	-- if 'Only target/focus' is on, check if target/focus guid is same as source
	if onlytarget and (UnitGUID("target") ~= src or UnitGUID("focus") ~= src) then return end
	
	local message
	if event == "used" then
		message = srcName.." "..L["USED"].." "..name
	elseif event == "spec" then
		message = srcName.." "..L["SPECCED"].." "..spell
	end

	if target == "screen" then
		RaidNotice_AddMessage(RaidWarningFrame, message, ChatTypeInfo["RAID_WARNING"])		
		PlaySoundFile("Sound\\Interface\\RaidWarning.wav")
	elseif target == "screen_nosound" then
		RaidNotice_AddMessage(RaidWarningFrame, message, ChatTypeInfo["RAID_WARNING"])		
	elseif target == "party" then
		SendChatMessage(message, "PARTY")
	elseif target == "raid" then
		SendChatMessage(message, "RAID")
	elseif target == "rw" then
		SendChatMessage(message, "RAID_WARNING")
	elseif target == "bg" then
		SendChatMessage(message, "BATTLEGROUND")
	else
		addon:Print(message)
	end
end

function addon:AlertSpec(source, spec)
	local _, class, _, _, _, name = GetPlayerInfoByGUID(source)
	local info = cdf:GetSpec(class, spec)
	if info then
		addon:SendAlert("spec", source, nil, info.name, "spec")
	end
end

function addon:PLAYER_ENTERING_WORLD()
	addon:ZONE_CHANGED_NEW_AREA()
end

-- fired when entering a new area, clean up and redraw everything
function addon:ZONE_CHANGED_NEW_AREA()
	--wipe unit data
	wipe(addon.Units)
	
	local db = addon.DB.profile
	local anchors = addon.Anchors
	local zonetype = select(2, IsInInstance())

	-- check if addon is enabled in options
	if db.Settings.EnableAddon then
		-- register events, redraw anchors
		for k in pairs(events) do
			addon:RegisterEvent(k)
		end
		addon:RefreshAnchors()
	else
		-- unregister events, clear and hide everything
		for k in pairs(events) do
			addon:UnregisterEvent(k)
		end
		for k, v in pairs(anchors) do
			cdf:ClearAnchor(v)
			v:Hide()
		end
	end
end

-- refresh target anchor when player target changes
function addon:PLAYER_TARGET_CHANGED()
	-- is anchor enabled?
	if addon:GetZoneSetting("Anchors", "Anchor_Target") == false then return end
	-- is it a valid enemy player?
	if UnitIsPlayer("target") and UnitIsEnemy("player", "target") then
		local GUID = UnitGUID("target")
		local spells = units[GUID]
		addon:UpdateAnchor(addon.Anchors.Anchor_Target, spells, GUID)
	else
		local hide = addon.DB.profile.Anchors.Anchor_Target.Autohide
		addon:UpdateAnchor(addon.Anchors.Anchor_Target, nil, hide)
	end
end

-- refresh focus anchor when player focus changes
function addon:PLAYER_FOCUS_CHANGED()
	-- is anchor enabled?
	if addon:GetZoneSetting("Anchors", "Anchor_Focus") == false then return end
	-- is it a valid enemy player?
	if UnitIsPlayer("focus") and UnitIsEnemy("player", "focus") then
		local GUID = UnitGUID("focus")
		local spells = units[GUID]
		addon:UpdateAnchor(addon.Anchors.Anchor_Focus, spells, GUID)
	else
		local hide = addon.DB.profile.Anchors.Anchor_Focus.Autohide
		addon:UpdateAnchor(addon.Anchors.Anchor_Focus, nil, hide)
	end
end

-- refresh arena anchors as opponent info updates
function addon:ARENA_OPPONENT_UPDATE()
	-- is anchor enabled?
	if addon.DB.profile.Anchors.Anchor_Arena.Enabled == false then return end
	local db = addon.DB.profile
	for i = 1, 5 do
		local unit = "arena"..i
		local anchor = "Anchor_Arena"..i
		if UnitIsPlayer(unit) then
			local GUID = UnitGUID(unit)
			local spells = units[GUID]
			addon:UpdateAnchor(addon.Anchors[anchor], spells, GUID)
		else
			addon:UpdateAnchor(addon.Anchors[anchor], nil, "hide")
		end
	end
end

-- watch spells
function addon:COMBAT_LOG_EVENT_UNFILTERED(event, timeStamp, eventType, hideCaster, srcGUID, srcName, srcFlags, srcRaidFlags, dstGUID, dstName, dstFlags, dstRaidFlags, ...)
	-- check event subtype
	if not eventtypes[eventType] then return end
	-- check if guid is valid and if unit is player and hostile
	if srcGUID and srcGUID ~= GUID_INVALID and not GetPlayerInfoByGUID(srcGUID) or not addon:IsHostile(srcFlags) then return end

	local spellID, spellName, spellSchool, auraType = ...
	
	-- some failchecks
	if spellName == "-1" or spellID == 0 then return end

	local db = addon.DB.profile
	if not units[srcGUID] then
		units[srcGUID] = {}
	end
	
	-- spec detection #1 (spammable spells)
	if addon.SpecSpells[spellID] and not units[srcGUID].spec then
		units[srcGUID].spec = addon.SpecSpells[spellID]
		addon:AlertSpec(srcGUID, spec)
		addon:RefreshAnchors()
	end
	
	-- check if spell is in database
	local spell = addon:GetSpell(spellID)
	if not spell or not spell.cooldown or spell.cooldown<1 then return end

	local spec = spell.spec or 0
	local glyphed = addon:GetGlyphSettings(spellID, spec)
	local spelltype = spell.type
	local anchors = addon.Anchors
	local cd = addon:GetSpellCooldown(spellID, glyphed, spec)
	
	-- spec detection #2 (cd spells)
	if spec and not units[srcGUID].spec then
		units[srcGUID].spec = spec
		addon:AlertSpec(srcGUID, spec)
		addon:RefreshAnchors()
	end
	
	if eventType == "SPELL_CAST_SUCCESS" then
		units[srcGUID][spellID] = GetTime()

		if addon.Anchors.Anchor_Target:IsShown() and srcGUID == UnitGUID("target") then
			cdf:UpdateSpell(anchors.Anchor_Target, spellID, cd, addon:GetBarColor("Anchor_Target", spelltype))
		end

		if addon.Anchors.Anchor_Focus:IsShown() and srcGUID == UnitGUID("focus") then
			cdf:UpdateSpell(anchors.Anchor_Focus, spellID, cd, addon:GetBarColor("Anchor_Focus", spelltype))
		end
		
		local barcolor = addon:GetBarColor("Anchor_Arena", spelltype)
		for i = 1, 5 do
			local a = "Anchor_Arena"..i
			local u = "arena"..i
			if addon.Anchors[a]:IsShown() and srcGUID == UnitGUID(u) then
				cdf:UpdateSpell(anchors[a], spellID, cd, barcolor)
			end		
		end

		addon:SendAlert("used", srcGUID, dstGUID, spellID, spelltype)
	end
end